from typing import Optional

from kubernetes import config, client
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name
import importlib
from rich.table import Table
from rich.console import Console

def get_current_tensorkube_version():
    return importlib.metadata.version('tensorkube')

def get_cluster_tensorkube_version(context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)
    try:
        config_map = v1.read_namespaced_config_map("tensorkube-migration", "default")
        return config_map.data['version']
    except client.exceptions.ApiException as e:
        if e.status == 404:
            return None
        else:
            raise e
        
def print_tensorkube_version():
    cli_version = get_current_tensorkube_version()
    cluster_version = get_cluster_tensorkube_version()
    console = Console()
    table = Table(title="TensorKube Versions")

    table.add_column("Component", justify="left", style="cyan", no_wrap=True)
    table.add_column("Version", justify="left", style="magenta")

    table.add_row("CLI Version", cli_version)
    table.add_row("Cluster Version", cluster_version)

    console.print(table)
    


import json
import os
import platform
import shutil
import subprocess

import click
import semver

from tensorkube.constants import CliColors, LOCKED_KUBECTL_VERSION, LOCKED_HELM_VERSION
from tensorkube.services.error import CLIVersionError


def check_and_install_kubectl():
    """Check if kubectl is installed and install it if it's not."""
    try:
        output = subprocess.run(["kubectl", "version", "--client", "-o", "json"], check=True, stdout=subprocess.PIPE)
        version_dict = json.loads(output.stdout.decode("utf-8"))
        version = semver.VersionInfo.parse(version_dict["clientVersion"]["gitVersion"].lstrip("v"))
        locked_kubectl_version = semver.VersionInfo.parse(LOCKED_KUBECTL_VERSION)
        print("kubectl is already installed.")
        if (version.major, version.minor) < (locked_kubectl_version.major, locked_kubectl_version.minor):
            text = f"kubectl version is {version}. Please upgrade kubectl to version above {LOCKED_KUBECTL_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("kubectl not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        if platform.system() == "Darwin" or platform.system() == "Linux":
            try:
                system_name = platform.system().lower()
                architecture = "arm64" if 'arm' in platform.machine() else "amd64"
                install_kubectl(system_name, architecture)
            except subprocess.CalledProcessError as e:
                print("Unable to install kubectl. Please install kubectl manually.")
                raise e
        else:
            print("Unsupported operating system. Please install kubectl manually.")
            raise Exception('Unsupported operating system.')

        # Verify kubectl installation
        try:
            subprocess.run("kubectl version --client", shell=True, check=True)
            print("kubectl installed successfully.")
        except subprocess.CalledProcessError as e:
            print("kubectl installation failed. Please install kubectl manually.")
            raise e


def check_and_install_helm():
    """Check if helm is installed and install it if it's not."""
    try:
        output = subprocess.run(["helm", "version", "--template={{.Version}}"], check=True, stdout=subprocess.PIPE)
        version = semver.VersionInfo.parse(output.stdout.decode("utf-8").strip().lstrip("v"))
        locked_helm_version = semver.VersionInfo.parse(LOCKED_HELM_VERSION)
        print("Helm is already installed.")
        if (version.major, version.minor) < (locked_helm_version.major, locked_helm_version.minor):
            text = f"Helm version is {version}. Please upgrade Helm to version above {LOCKED_HELM_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(click.style("Helm not found. Proceeding with installation. Might require sudo password.", bold=True,
                               fg=CliColors.WARNING.value))
        if platform.system() == "Darwin" or platform.system() == "Linux":
            try:
                system_name = platform.system().lower()
                architecture = "arm64" if 'arm' in platform.machine() else "amd64"
                install_helm_using_downloader(system_name=system_name, architecture=architecture)
            except Exception as e:
                click.echo(f"Error while installing helm: {e}. Please install helm manually", err=True)
                raise e
        else:
            print("Unsupported operating system. Please install helm manually.")
            raise Exception('Unsupported operating system.')

        # Verify helm installation
        try:
            subprocess.run(["helm", "version"], check=True)
            print("Helm installed successfully.")
        except subprocess.CalledProcessError as e:
            print("Helm installation failed. Please install helm manually.")
            raise e


def install_helm_using_downloader(system_name, architecture):
    try:
        # Download Helm installer script
        click.echo("Downloading Helm installer script...")

        # Install Helm by downloading specific version
        click.echo(f"Downloading Helm {LOCKED_HELM_VERSION} for {system_name}-{architecture} architecture...")

        download_url = f"https://get.helm.sh/helm-v{LOCKED_HELM_VERSION}-{system_name}-{architecture}.tar.gz"
        tarball_name = f"helm-{LOCKED_HELM_VERSION}-{system_name}-{architecture}.tar.gz"
        subprocess.run(f"curl -fsSL {download_url} -o {tarball_name}", shell=True, check=True)

        click.echo("Helm tarball downloaded successfully.")

        # Extract the tarball
        click.echo("Extracting Helm tarball...")
        subprocess.run(f"tar -zxvf {tarball_name}", shell=True, check=True)

        # Move the helm binary to /usr/local/bin
        click.echo("Moving Helm binary to /usr/local/bin...")
        subprocess.run(f"sudo mv {system_name}-{architecture}/helm /usr/local/bin/", shell=True, check=True)

        # Clean up the downloaded tarball and extracted files
        click.echo("Cleaning up downloaded files...")
        os.remove(tarball_name)
        shutil.rmtree(f"{system_name}-{architecture}")
        subprocess.run(f"rm -rf helm-{LOCKED_HELM_VERSION}-{system_name}-{architecture}", shell=True, check=True)

    except subprocess.CalledProcessError as e:
        click.echo(f"An error occurred: {e}", err=True)
        raise e
    except OSError as e:
        click.echo(f"Error deleting the installer script: {e}", err=True)


def install_kubectl(system, architecture):
    valid_architectures = {"amd64", "arm64"}
    if architecture not in valid_architectures:
        click.echo(f"Invalid architecture: {architecture}. Valid options are: {valid_architectures}", err=True)
        return

    try:
        # Download kubectl binary for the specified architecture
        click.echo(f"Downloading kubectl binary for {architecture} architecture...")
        cmd = f"curl -LO https://dl.k8s.io/release/v{LOCKED_KUBECTL_VERSION}/bin/{system}/{architecture}/kubectl"
        subprocess.run(
            cmd,
            shell=True, check=True)
        click.echo(f"kubectl binary for {architecture} downloaded successfully.")

        # Make the binary executable
        subprocess.run("chmod +x kubectl", shell=True, check=True)

        # Move it to /usr/local/bin
        subprocess.run("sudo mv kubectl /usr/local/bin/", shell=True, check=True)

        click.echo(f"kubectl for {architecture} installed successfully.")

    except subprocess.CalledProcessError as e:
        click.echo(f"An error occurred: {e}", err=True)



import click
import yaml
import os
import re
from typing import List, Dict, Optional

from tensorkube.constants import CliColors
from tensorkube.services.volumes_service import SUPPORTED_VOLUME_TYPES, check_existing_volume_pv_and_pvc, \
    create_volume_k8s_resources, get_volume_cloud_resource


class Config(object):
    def __init__(self, config_file_path):
        self.__config = {}
        if config_file_path != None:
            file_path = os.path.abspath(config_file_path)  # Convert to absolute path
            if not os.path.isfile(file_path):
                raise Exception(f"Config.yaml file not found in path: {file_path}")
            with open(file_path, 'r') as config_file:
                self.__config = yaml.safe_load(config_file)

    def get(self, key):
        return self.__config.get(key, None)

    def set_if_not_exist(self, key, value):
        if key in self.__config:
            return self
        if isinstance(value, tuple):
            self.__config[key] = list(value)
        else:
            self.__config[key] = value
        return self

    def __str__(self) -> str:
        return str(self.__config)


def run_deployment_config_pre_flight_checks(config: Config) -> bool:
    return run_volume_checks(volumes=config.get("volumes"), env=config.get("env"))

def is_valid_mount_path(mount_path: Optional[str]):
    if not mount_path:
        return True
    pattern = r'^(\/[a-zA-Z0-9._-]+)+\/?$'
    return bool(re.match(pattern, mount_path))

def run_volume_checks(volumes: Optional[List[Dict]], env: Optional[str]) -> bool:
    if not volumes:
        return True

    for volume in volumes:
        volume_name = volume.get("name", None)
        volume_type = volume.get("type", None)
        volume_mount_path = volume.get("mount_path", None)

        if not volume_type or not volume_name:
            click.echo(click.style("Volume name and type are required", fg=CliColors.ERROR.value))
            return False

        if volume_type not in SUPPORTED_VOLUME_TYPES:
            click.echo(click.style(f"Volume type '{volume_type}' is not supported", fg=CliColors.ERROR.value))
            return False

        if not is_valid_mount_path(volume_mount_path):
            click.echo(click.style(f"Mount path '{volume_mount_path}' for volume '{volume_name}' is invalid. Refer to our docs for mount path guidelines: https://tensorfuse.io/docs/concepts/volumes", fg=CliColors.ERROR.value))
            return False

        click.echo(f"Checking volume '{volume_name}' of type '{volume_type}'")
        cloud_volume_resource = get_volume_cloud_resource(volume_name, volume_type)
        if not cloud_volume_resource:
            click.echo(click.style(f"Volume '{volume_name}' of type '{volume_type}' not found", fg=CliColors.ERROR.value))
            return False

        click.echo(f"Cloud resources for volume '{volume_name}' of type '{volume_type}' found")
        is_existing_volume_pv_and_pvc = check_existing_volume_pv_and_pvc(volume_name, volume_type, env)
        if not is_existing_volume_pv_and_pvc:
            try:

                click.echo(f"Not Found: Cluster resources for volume '{volume_name}' of type '{volume_type}'")
                click.echo(f"Creating cluster resources...")
                create_volume_k8s_resources(volume_name=volume_name, volume_type=volume_type,
                                            cloud_volume_resource=cloud_volume_resource , env=env)
            except Exception as e:
                click.echo(click.style(f"Error creating volume resources: {e}", fg=CliColors.ERROR.value))
                return False
        else:
            click.echo(f"Cluster resources for volume '{volume_name}' of type '{volume_type}' found")

    return True



from tensorkube.services.job_queue_service import create_cloud_resources_for_queued_job_support
from tensorkube.services.aws_service import get_bucket_name, get_credentials
from tensorkube.services.s3_service import create_s3_bucket
from tensorkube.services.k8s_service import create_aws_secret
from tensorkube.constants import get_cluster_name
from tensorkube.services.iam_service import attach_role_policy, get_aws_account_id
import click

def apply():
    try:
        bucket_name = get_bucket_name(env_name='keda', type='train')
        create_s3_bucket(bucket_name)
        cluster_name = get_cluster_name()
        role_name = f"{cluster_name}-sqs-access-role"
        dyanmo_policy_name = f"{cluster_name}-dynamo-access-policy"
        attach_role_policy(get_aws_account_id(), dyanmo_policy_name, role_name)
        create_aws_secret(get_credentials(),"keda")
        click.echo("Successfully Created Cloud Resources for Queued Job Support")
    except Exception as e:
        raise e


from datetime import datetime
from typing import Optional, List
from kubernetes import config, utils
from pkg_resources import resource_filename
import yaml
import click

from tensorkube.constants import DEFAULT_NAMESPACE, GPU_DEPLOYMENT_MIN_MEMORY, CliColors, GPU_DEPLOYMENT_MIN_CPU, \
    PodStatus
from tensorkube.services.deploy import DEFAULT_ENTRYPOINT, DEFAULT_GITHUB_ACTIONS, DEFAULT_ENV
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name, get_pod_name_corresponing_to_job, \
    start_streaming_pod, find_and_delete_old_job
from tensorkube.services.knative_service import get_instance_family_from_gpu_type, apply_anti_affinity_for_gpus
from tensorkube.services.nodepool import get_gpu_nodepool
from tensorkube.services.nydus import get_nydus_runtime_class_name


DEFAULT_FOLLOW = False


def get_one_shot_job_name(sanitised_project_name: str) -> str:
    return f"{sanitised_project_name}-one-shot-job"


def deploy_one_shot_job(name: Optional[str], sanitised_project_name: str, image_url: str,
                        gpus: int, gpu_type: str, cpu: float = 100, memory: int = 200, env: Optional[str] = DEFAULT_ENV,
                        entrypoint: Optional[str] = DEFAULT_ENTRYPOINT, secrets: List[str] = None,
                        context_name: Optional[str] = None, follow: bool = DEFAULT_FOLLOW,
                        github_actions: bool = DEFAULT_GITHUB_ACTIONS) -> None:
    if secrets is None:
        secrets = []
    namespace = env if env else DEFAULT_NAMESPACE

    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)

    # Read the YAML file
    yaml_file_path = resource_filename('tensorkube', 'configurations/build_configs/one_shot_job.yaml')
    with open(yaml_file_path, 'r') as f:
        yaml_content = f.read()

    job_name = name if name else get_one_shot_job_name(sanitised_project_name)
    click.echo(f"Creating One-Shot Job `{job_name}` for project {sanitised_project_name} and {gpus} GPUs.")
    yaml_content = yaml_content.replace('${JOB_NAME}', job_name)
    yaml_content = yaml_content.replace('${GPUS}', str(gpus))

    # Load the YAML content
    yaml_dict = yaml.safe_load(yaml_content)

    yaml_dict['metadata']['namespace'] = namespace
    yaml_dict['spec']['template']['spec']['containers'][0]['image'] = image_url
    yaml_dict['spec']['template']['spec']['containers'][0]['env'][0]['value'] = "us-east-1"

    if secrets:
        yaml_dict['spec']['template']['spec']['containers'][0]['envFrom'] = [{
            'secretRef': {
                'name': secret_name
            }

        } for secret_name in secrets]

    if gpus > 0:
        yaml_dict['spec']['template']['spec']['runtimeClassName'] = get_nydus_runtime_class_name()
        if 'nodeSelector' in yaml_dict['spec']['template']['spec']:
            yaml_dict['spec']['template']['spec']['nodeSelector']['karpenter.sh/nodepool'] = get_gpu_nodepool()
        else:
            yaml_dict['spec']['template']['spec']['nodeSelector'] = {'karpenter.sh/nodepool': get_gpu_nodepool()}

    if gpus > 0:
        # by default gpu pods usually use all of available memory and cpu
        # so I am here defining some limits above which the cpu and memory parameters will be used
        if memory >= GPU_DEPLOYMENT_MIN_MEMORY:
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'memory'] = f'{str(int(memory))}M'
        else:
            click.echo(click.style(
                f"Memory is not specified or is less than the minimum required for GPU deployment. Setting to minimum: {GPU_DEPLOYMENT_MIN_MEMORY}M.",
                fg=CliColors.WARNING.value))
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'memory'] = f'{str(GPU_DEPLOYMENT_MIN_MEMORY)}M'
        if cpu >= GPU_DEPLOYMENT_MIN_CPU:
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'cpu'] = f'{str(int(cpu))}m'
        else:
            click.echo(click.style(
                f"CPU is not specified or is less than the minimum required for GPU deployment. Setting to minimum: {GPU_DEPLOYMENT_MIN_CPU}m.",
                fg=CliColors.WARNING.value))
            yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
                'cpu'] = f'{str(GPU_DEPLOYMENT_MIN_CPU)}m'

        if 'nodeSelector' in yaml_dict['spec']['template']['spec']:
            yaml_dict['spec']['template']['spec']['nodeSelector'][
                'karpenter.k8s.aws/instance-family'] = get_instance_family_from_gpu_type(gpu_type)
        else:
            yaml_dict['spec']['template']['spec']['nodeSelector'] = {
                'karpenter.k8s.aws/instance-family': get_instance_family_from_gpu_type(gpu_type)}

        if yaml_dict['spec']['template']['spec']['containers'][0].get('volumeMounts', None):
            yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'].append(
                {'name': 'dshm', 'mountPath': '/dev/shm'})
            yaml_dict['spec']['template']['spec']['volumes'].append(
                {'name': 'dshm', 'emptyDir': {'medium': 'Memory', 'sizeLimit': '10Gi'}})
        else:
            yaml_dict['spec']['template']['spec']['containers'][0]['volumeMounts'] = [{'name': 'dshm', 'mountPath': '/dev/shm'}]
            yaml_dict['spec']['template']['spec']['volumes'] = [{'name': 'dshm', 'emptyDir': {'medium': 'Memory', 'sizeLimit': '10Gi'}}]
    else:
        yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests'][
            'memory'] = f'{str(int(memory))}M'
        yaml_dict['spec']['template']['spec']['containers'][0]['resources']['requests']['cpu'] = f'{str(int(cpu))}m'
        apply_anti_affinity_for_gpus(yaml_dict)


    yaml_dict['spec']['template']['metadata'] = {
        'annotations': {'deploy_time': datetime.now().isoformat()}}

    if entrypoint:
        yaml_dict['spec']['template']['spec']['containers'][0]['command'] = [entrypoint]

    old_job_deleted = find_and_delete_old_job(job_name=job_name, namespace=namespace)
    if not old_job_deleted:
        click.echo("Another Job is already in progress. Please wait for it to complete.")
        if github_actions:
            raise Exception("Another Job is already in progress. Please wait for it to complete.")
        return

    utils.create_from_dict(k8s_api_client, yaml_dict)
    click.echo('Deployed your one shot job.')

    if follow:
        job_pod_name = get_pod_name_corresponing_to_job(job_name=job_name, namespace=namespace)
        if job_pod_name is None:
            click.echo("Job pod not found")
            return PodStatus.FAILED.value
        # TODO: stream multiple lines instead of one by one
        start_streaming_pod(pod_name=job_pod_name, namespace=namespace, status=PodStatus.SUCCEEDED)







import click

from tensorkube.constants import TENSORFUSE_NAMESPACES, DEFAULT_NAMESPACE, get_cluster_name, \
    get_efs_security_group_name, get_mount_policy_name, get_mount_driver_role_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.services.filesystem_service import configure_efs_for_the_cluster, apply_efs_pv, apply_efs_pvc, \
    get_efs_pvc_name, get_efs_pv_name, delete_efs_filesystem, get_efs_name, delete_security_group
from tensorkube.services.iam_service import create_mountpoint_iam_policy, attach_role_policy, detach_role_policy, \
    delete_policy
from tensorkube.services.k8s_service import create_new_namespace, list_all_namespaces, delete_namespace, \
    create_build_pv_and_pvc, delete_pvc_using_name_and_namespace, get_s3_claim_name, delete_pv_using_name, \
    get_s3_pv_name, delete_all_jobs_in_namespace, delete_aws_secret
from tensorkube.services.knative_service import delete_all_ksvc_from_namespace
from tensorkube.services.s3_service import create_s3_bucket, delete_s3_bucket, get_bucket_name
from tensorkube.services.s3_access_service import delete_s3_access_to_pods
from tensorkube.services.build import configure_buildkit_irsa


def create_new_environment(env_name: str):
    # create a new namespace for the environment
    # create an s3 bucket for the environment
    # Create NFS mounts for the environment
    if env_name is None:
        click.echo("Please provide a valid environment name")
        return
    if env_name in list_environments():
        click.echo("Environment already exists. Skipping creation")
        return
    if env_name in TENSORFUSE_NAMESPACES or env_name == DEFAULT_NAMESPACE:
        click.echo("Cannot create a system namespace")
        return
    # Create S3 Bucket
    bucket_name = get_bucket_name(env_name=env_name)
    cluster_name = get_cluster_name()
    create_s3_bucket(bucket_name)

    create_new_namespace(env_name=env_name)
    
    policy_name = get_mount_policy_name(cluster_name=cluster_name, env=env_name)
    create_mountpoint_iam_policy(policy_name=policy_name, bucket_name=bucket_name)
    attach_role_policy(account_no=get_aws_account_id(), policy_name=policy_name,
                       role_name=get_mount_driver_role_name(cluster_name=cluster_name))
    # Create PV and PVC for s3
    create_build_pv_and_pvc(bucket_name, env=env_name)
    # configure efs for the environment
    configure_efs_for_the_cluster(env=env_name)
    apply_efs_pv(env=env_name)
    apply_efs_pvc(env=env_name)
    configure_buildkit_irsa(env_name=env_name)


def delete_environment(env_name: str, is_teardown = False):
    # delete the namespace for the environment
    # delete the s3 bucket for the environment
    # delete the NFS mounts for the environment
    if env_name is None:
        click.echo("Please provide a valid environment name to delete")
        return
    if env_name in TENSORFUSE_NAMESPACES:
        click.echo("Cannot delete a system namespace")
        return
    if env_name == DEFAULT_NAMESPACE and (not is_teardown):
        click.echo("Cannot delete a system namespace")
        return
    if env_name not in list_environments():
        click.echo("Environment does not exist. Skipping deletion")
        return
    # delete all ksvcs in the namespace
    delete_all_ksvc_from_namespace(namespace=env_name)
    delete_all_jobs_in_namespace(namespace=env_name)
    # delete all s3 configuration
    bucket_name = get_bucket_name(env_name=env_name)
    cluster_name = get_cluster_name()
    policy_name = get_mount_policy_name(cluster_name=cluster_name, env=env_name)
    try:
        detach_role_policy(account_no=get_aws_account_id(), policy_name=policy_name,
                           role_name=get_mount_driver_role_name(cluster_name=cluster_name))
    except Exception as e:
        click.echo(f"An error occurred while detaching policy from role: {e}")

    try:
        delete_policy(get_aws_account_id(), policy_name)
    except Exception as e:
        click.echo(f"An error occurred while deleting policy: {e}")
    delete_pvc_using_name_and_namespace(pvc_name=get_s3_claim_name(env_name=env_name), namespace=env_name)
    delete_pv_using_name(pv_name=get_s3_pv_name(env_name=env_name))
    delete_s3_bucket(bucket_name=bucket_name)
    train_bucket_name = get_bucket_name(env_name=env_name, type='train')
    delete_s3_bucket(bucket_name=train_bucket_name)
    delete_aws_secret(env_name)
    delete_s3_access_to_pods(env_name)
    # delete efs
    delete_pvc_using_name_and_namespace(pvc_name=get_efs_pvc_name(env=env_name), namespace=env_name)
    delete_pv_using_name(pv_name=get_efs_pv_name(env=env_name))
    delete_efs_filesystem(get_efs_name(cluster_name=get_cluster_name(), env=env_name))
    delete_security_group(get_efs_security_group_name(env=env_name))
    # delete the namespace
    delete_namespace(env_name)


def list_environments():
    namespaces = list_all_namespaces()
    environments = []
    for namespace in namespaces:
        if namespace not in TENSORFUSE_NAMESPACES:
            environments.append(namespace)
    return environments


def check_environment_exists(env_name: str) -> bool:
    return env_name in list_environments()



# Step 1: Ask the domain from the user and store it in a variable. Double confirm the domain from the user before setting up things
# Step 2: Check if a tensorkube certificate exists for the same domain. If it does, add the domain to the existing certificate.
# Step 3: If the certificate does not exist, create a new certificate for the domain.
# STep 4: Create a ClusterdomainClaim
# STep 5: Create a Cluster
import re
from dataclasses import dataclass
from typing import Optional, List, Dict

import click
from kubernetes import config, client
from pkg_resources import resource_filename

from tensorkube.constants import get_cluster_name
from tensorkube.services.aws_service import get_cloudformation_client
from tensorkube.services.cloudformation_service import deploy_generic_cloudformation_stack, \
    get_stack_status_from_stack_name, CfnStackStatus
from tensorkube.services.istio import configure_ssl_for_ingress_gateway
from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name


@dataclass
class ServiceDomains:
    service_name: str
    domains: List[str]
    namespace: str



def get_all_virtualservices(namespaces: List[str]) -> List[Dict]:
    """
    Fetch all VirtualServices from a list of namespaces.

    Args:
        namespaces (List[str]): List of namespaces to query for VirtualServices.

    Returns:
        List[Dict]: A list of VirtualServices from all specified namespaces.
    """
    try:
        # Load Kubernetes configuration
        config.load_kube_config()
        k8s_client = client.CustomObjectsApi()

        # Initialize an empty list to store VirtualServices
        virtualservices = []

        # Fetch VirtualServices from each namespace
        for namespace in namespaces:
            vs_list = k8s_client.list_namespaced_custom_object(
                group="networking.istio.io",
                version="v1",
                namespace=namespace,
                plural="virtualservices"
            )
            virtualservices.extend(vs_list.get('items', []))

        return virtualservices

    except Exception as e:
        raise Exception(f"Failed to fetch VirtualServices: {e}")


def find_domains_for_service(virtualservices: List[Dict], service_name: str, namespace: str) -> ServiceDomains:
    """Process VirtualServices to find domains for a specific service"""
    domains = []
    for vs in virtualservices:
        vs_namespace = vs.get('metadata', {}).get('namespace')
        if vs_namespace != namespace:
            continue  # Skip if it's not in the same namespace
        for http in vs['spec'].get('http', []):
            for route in http.get('route', []):
                if route.get('destination', {}).get('host', '').startswith(service_name):
                    domains.extend(vs['spec'].get('hosts', []))

    return ServiceDomains(
        service_name=service_name,
        domains=list(set(domains)),
        namespace=namespace,
        # Remove duplicates
    )


def get_certificate_stack_name(domain: str) -> str:
    """
    Generate the CloudFormation stack name for the certificate.
    """
    return f"{get_cluster_name()}-cert-{domain.replace('.', '-')}"


def validate_domain(domain: str) -> bool:
    """
    Validates the domain format.
    :param domain: Domain name to validate
    :return: True if valid, False otherwise
    """
    pattern = re.compile(r'^(?:[a-zA-Z0-9-]{1,63}\.)+(?:[a-zA-Z]{2,})$')
    return bool(pattern.match(domain))


def deploy_certificate_stack(domain: str) -> bool:
    """
    Deploy ACM certificate stack using the generic deployment function
    """
    if not validate_domain(domain):
        click.echo(
            click.style(f'Invalid domain format: {domain}', fg='red')
        )
        return False
    stack_name = get_certificate_stack_name(domain)
    cluster_name = get_cluster_name()

    # Load certificate template
    template_file = resource_filename('tensorkube', 'configurations/cloudformation/certificate_manager.yaml')
    with open(template_file) as file:
        template_body = file.read()

    parameters = [{"ParameterKey": "DomainName", "ParameterValue": domain},
                  {"ParameterKey": "ClusterName", "ParameterValue": cluster_name}]

    queued = deploy_generic_cloudformation_stack(stack_name=stack_name, template_body=template_body,
                                                 parameters=parameters, capabilities=["CAPABILITY_NAMED_IAM"],
                                                 should_wait=False)
    if queued:
        click.echo(f'Certificate queued. Getting your validation records ready. Please wait.')
    else:
        click.echo(f'Failed to queue certificate.')
    return queued


def configure_certificate(domain):
    """Deploy ACM certificate for domain."""
    if not validate_domain(domain):
        click.echo(f"Invalid domain format: {domain}")
        return

    click.echo(f"Deploying certificate for *.{domain}")
    result = deploy_certificate_stack(domain)
    return result


def get_existing_ssl_config(namespace: str = 'istio-system', service_name: str = 'istio-ingressgateway') -> Optional[
    dict]:
    """
    Check if SSL annotations exist and get certificate ARN
    Returns:
        Optional[dict]: Dictionary with certificate_arn and ssl_ports if found, None otherwise
    """
    try:
        # Load kubernetes config
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
        api_client = config.new_client_from_config(context=context_name)
        v1 = client.CoreV1Api(api_client)

        # Get existing service
        service = v1.read_namespaced_service(name=service_name, namespace=namespace)

        # Check for annotations
        annotations = service.metadata.annotations or {}
        cert_annotation = 'service.beta.kubernetes.io/aws-load-balancer-ssl-cert'
        ports_annotation = 'service.beta.kubernetes.io/aws-load-balancer-ssl-ports'

        if cert_annotation in annotations:
            return {'certificate_arn': annotations[cert_annotation],
                    'ssl_ports': annotations.get(ports_annotation, '443')}
        return None

    except client.exceptions.ApiException as e:
        click.echo(f"Failed to get SSL config: {str(e)}")
        return None
    except Exception as e:
        click.echo(f"Unexpected error getting SSL config: {str(e)}")
        return None


def get_certificate_arn_from_stack(domain: str) -> Optional[str]:
    """Get certificate ARN from CloudFormation stack outputs"""
    stack_name = get_certificate_stack_name(domain=domain)
    cfn_client = get_cloudformation_client()

    try:
        response = cfn_client.describe_stacks(StackName=stack_name)
        outputs = response['Stacks'][0]['Outputs']
        cert_arn = next(output['OutputValue'] for output in outputs if output['OutputKey'] == 'CertificateArn')
        return cert_arn
    except Exception as e:
        click.echo(f"Error getting certificate ARN: {str(e)}")
        return None

def check_if_a_domain_is_validated(domain:str):
    # extract the base domain from the subdomain
    # check if the cloudformation is created
    # check if the domain is attached to the load balancer
    base_domain = '.'.join(domain.split('.')[1:])
    stack_name = get_certificate_stack_name(domain=base_domain)
    stack_status = get_stack_status_from_stack_name(stack_name)
    if stack_status != CfnStackStatus.CREATE_COMPLETE:
        click.echo(
            click.style(
                "No such domain configured or validated. Please run `tensorkube configure domain` to configure a domain",
                fg="red"
            )
        )
        return False
    cert_arn = get_certificate_arn_from_stack(domain=base_domain)
    existing_config = get_existing_ssl_config()
    if existing_config is None:
        old_cert_arn = None
    else:
        old_cert_arn = existing_config["certificate_arn"]
    if old_cert_arn != cert_arn:
        return False
    return True


def attach_domain_name(domain_name, force: bool = False):
    # Check if the cloudformation stack for this name exists and is in successful state
    # if does note exist ask to create
    # if exists but not ready - ask to complete validation using validation command
    # if exists and ready check if the cluster has some other certificate arn configured
    # if yes then get the domain from the certificate arn and display that this has already been configured
    # with the following domain - and ask it to use the --force method to override
    # if no then get the certificate arn from the stack and attach it to the cluster
    # after attaching check if it has been successfully attached and display the message
    stack_name = get_certificate_stack_name(domain_name)
    print(f"Stack name : {stack_name}")
    stack_status = get_stack_status_from_stack_name(stack_name)
    print(f"Stack status : {stack_status}")
    if stack_status is None or (
            stack_status != CfnStackStatus.CREATE_IN_PROGRESS and stack_status != CfnStackStatus.CREATE_COMPLETE):
        click.echo(
            click.style("No such domain validated. Please wait for DNS validation"
                        " or create new mapping using `tensorkube domain create` command.",
                        fg="red"))
        return
    if stack_status == CfnStackStatus.CREATE_IN_PROGRESS:
        click.echo(click.style(
            "Domain validation pending. Please get validation details using `tensorkube domain get-validation-records` command"
            "or wait for your DNS provider to finish validating once this is done.",
            "yellow"))
    if stack_status == CfnStackStatus.CREATE_COMPLETE:
        new_cert_arn = get_certificate_arn_from_stack(domain=domain_name)
        existing_config = get_existing_ssl_config()
        if existing_config is None:
            old_cert_arn = None
        else:
            old_cert_arn = existing_config["certificate_arn"]
        print(f"Old cert arn: {old_cert_arn}")
        print(f"New cert arn: {new_cert_arn}")
        if old_cert_arn == new_cert_arn:
            print("Certificate already attached")
            configure_ssl_for_ingress_gateway(new_cert_arn)
        if old_cert_arn is not None and old_cert_arn != new_cert_arn:
            if not force:
                click.echo(click.style(
                    "Configuring multiple domains for the same cluster is not supported.", "red"))
            else:
                click.echo(click.style(
                    "Configuring multiple domains for the same cluster is not supported.", "red"))
        elif old_cert_arn is None:
            print("Attaching certificate to the cluster")
            configure_ssl_for_ingress_gateway(new_cert_arn)

def is_custom_domain(domain: str, service_env:str) -> bool:
    # Helper function to determine if a domain is custom/not generated by Knative or sslip.io
    return not (domain.endswith('.sslip.io') or 'svc.cluster.local' in domain or 'default.svc' in domain or domain.endswith('.default')
                or domain.endswith(f'.{service_env}') or domain.endswith(f'{service_env}.svc'))



from tensorkube.constants import NODEPOOL, EC2NODECLASS, GPU_NODEPOOL_VERSION

def gpu_nodepool_version():
    return GPU_NODEPOOL_VERSION

def get_gpu_nodepool():
    return f"{gpu_nodepool_version()}-nodepool"

def get_gpu_ec2nodeclass():
    return f"{gpu_nodepool_version()}-ec2nodeclass"

def get_config_file(type: str, name: str):
    if type != NODEPOOL and type != EC2NODECLASS:
        return None
    if name == 'default':
        return f"karpenter_{type}.yaml"
    name = name.replace("-", "_")
    return f"karpenter_{name}_{type}.yaml"





import click
import subprocess
from pkg_resources import resource_filename
from tensorkube.constants import get_cluster_name
from tensorkube.services.configure_service import update_cfn_configure_stack

def apply():
    try:
        folder_name = "configurations/cloudformation/configure/cluster_access_lambda_files/helm-charts/tk-karpenter-config"
        chart_path = resource_filename("tensorkube", folder_name)
        release_name = "karpenter-configuration"
        subprocess.run(["helm", "upgrade", "--install", release_name,  chart_path, "--set", f"clusterName={get_cluster_name()}"], check=True)
        updated_params = {
            "TemplatesVersion": "v0.0.2",
            "AWSAccessLambdaFunctionImageVersion": "v1.0.1",
            "CliVersion": "0.0.80",
            "EksAccessLambdaFunctionImageVersion": "v1.0.1"
        }
        update_cfn_configure_stack(updated_parameters=updated_params, test=False)

    except Exception as e:
        click.echo(click.style(f"Failed to apply karpenter configuration helm chart from {chart_path}: {e}", bold=True, fg="red"))
        raise e


from tensorkube.services.karpenter_service import upgrade_karpenter_nodepools
import click
def apply():
    try:
        upgrade_karpenter_nodepools()
        click.echo("Successfully upgraded Karpenter nodepools to L40")
    except Exception as e:
        raise e

    




import click

from tensorkube.constants import get_cluster_name, ADDON_NAME, get_mount_policy_name, get_mount_driver_role_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.services.build import delete_buildkit_irsa
from tensorkube.services.cloudformation_service import delete_cloudformation_stack, delete_launch_templates, \
    delete_tensorkube_base_stack
from tensorkube.services.eks_service import delete_eks_addon, delete_knative_core, delete_knative_crds, \
    delete_karpenter_from_cluster
from tensorkube.services.eksctl_service import delete_cluster
from tensorkube.services.environment_service import list_environments, delete_environment
from tensorkube.services.filesystem_service import cleanup_filesystem_resources
from tensorkube.services.iam_service import detach_role_policy, delete_role, delete_policy
from tensorkube.services.istio import remove_domain_server, uninstall_istio_from_cluster
from tensorkube.services.job_queue_service import teardown_job_queue_support
from tensorkube.services.knative_service import delete_knative_services, cleanup_knative_resources
from tensorkube.services.logging_service import teardown_cloudwatch
from tensorkube.services.nydus import delete_nydus
from tensorkube.services.s3_service import delete_s3_bucket, get_bucket_name



def cfn_teardown():
    delete_tensorkube_base_stack()


def legacy_teardown():
    click.echo("Deleting all job queue resources...")
    try:
        teardown_job_queue_support()
    except Exception as e:
        click.echo(f"Error while deleting job queue resources: {e}")

    # delete all services
    try:
        delete_knative_services()
    except Exception as e:
        click.echo("Error while deleting Knative services.")
    try:
        cleanup_filesystem_resources()
    except Exception as e:
        click.echo(f"Error while cleaning up filesystem resources: {e}")

    # EKS addon
    try:
        click.echo("Deleting EKS addon...")
        delete_eks_addon(get_cluster_name(), ADDON_NAME)
    except Exception as e:
        click.echo(f"Error while deleting EKS addon: {e}")

    # teardown cloudwatch
    try:
        teardown_cloudwatch()
    except Exception as e:
        click.echo(f"Error while tearing down Cloudwatch: {e}")

    click.echo("Deleting Enviroments...")
    try:
        environments = list_environments()
        for env in environments:
            click.echo(f"Deleting environment: {env}")
            delete_environment(env_name=env)
    except Exception as e:
        click.echo(f"Error while deleting environments: {e}")

    # Detach policy from role, delete role, delete policy
    click.echo("Deleting mountpoint driver role and policy...")
    click.echo("Detaching policy from role...")
    try:
        detach_role_policy(get_aws_account_id(), get_mount_driver_role_name(get_cluster_name()),
                           get_mount_policy_name(get_cluster_name()))
        click.echo("Deleting role...")
        delete_role(get_mount_driver_role_name(get_cluster_name()))
        click.echo("Deleting policy...")
        delete_policy(get_aws_account_id(), get_mount_policy_name(get_cluster_name()))
    except Exception as e:
        click.echo(f"Error while deleting role and policy: {e}")

    # delete s3 bucket
    click.echo("Deleting S3 bucket...")
    try:
        delete_s3_bucket(get_bucket_name())
    except Exception as e:
        click.echo(f"Error while deleting S3 bucket: {e}")

    click.echo("Deleting buildkit irsa...")
    try:
        delete_buildkit_irsa()
    except Exception as e:
        click.echo(f"Error while deleting buildkit irsa: {e}")

    click.echo("Uninstalling nydus helm charts...")
    try:
        delete_nydus()
    except Exception as e:
        click.echo(f"Error while uninstalling nydus resources: {e}")

    click.echo("Uninstalling domain server...")
    try:
        remove_domain_server()
    except Exception as e:
        click.echo(f"Error while uninstalling domain server: {e}")

    click.echo("Uninstalling Knative resources")
    try:
        cleanup_knative_resources()
    except Exception as e:
        click.echo(f"Error while cleaning up Knative resources: {e}")

    click.echo("Uninstalling and deleting Istio resources")
    try:
        uninstall_istio_from_cluster()
    except Exception as e:
        click.echo(f"Error while uninstalling Istio: {e}")
    click.echo("Uninstalling Knative core")
    try:
        delete_knative_core()
        click.echo("Uninstalling Knative CRDs")
        delete_knative_crds()
        click.echo("Successfully uninstalled Knative and Istio.")
    except Exception as e:
        click.echo(f"Error while uninstalling Knative: {e}")

    # remove karpenter
    click.echo("Uninstalling Karpenter...")
    try:
        delete_karpenter_from_cluster()
        click.echo("Successfully uninstalled Karpenter.")
    except Exception as e:
        click.echo(f"Error while uninstalling Karpenter: {e}")
    # delete cluster
    try:
        click.echo("Deleting cluster...")
        delete_cluster()
        click.echo("Successfully deleted cluster.")
    except Exception as e:
        click.echo(f"Error while deleting cluster.: {e}")
    try:
        # delete cloudformation stack
        click.echo("Deleting cloudformation stack...")
        delete_cloudformation_stack(get_cluster_name())
        click.echo("Successfully deleted cloudformation stack.")
    except Exception as e:
        click.echo(f"Error while deleting cloudformation stack: {e}")

    # delete launch templates
    click.echo("Deleting launch templates...")
    delete_launch_templates()


KNATIVE_VERSION = '1.15.0'

DOMAIN_SERVER_URL = 'https://github.com/knative/serving/releases/download/knative-v1.14.0/serving-default-domain.yaml' # not in use anymore
KNATIVE_ISTIO_CONTROLLER_URL = f'https://github.com/knative/net-istio/releases/download/knative-v{KNATIVE_VERSION}/net-istio.yaml'
KNATIVE_CRD_URL = f'https://github.com/knative/serving/releases/download/knative-v{KNATIVE_VERSION}/serving-crds.yaml'
KNATIVE_CORE_URL = f'https://github.com/knative/serving/releases/download/knative-v{KNATIVE_VERSION}/serving-core.yaml'




from tensorkube.services.nodepool import get_gpu_ec2nodeclass, get_config_file, gpu_nodepool_version
from tensorkube.constants import EC2NODECLASS
from tensorkube.services.karpenter_service import apply_ec2nodeclass
import click

def apply():
    try:
        apply_ec2nodeclass(get_config_file(EC2NODECLASS, gpu_nodepool_version()), get_gpu_ec2nodeclass(), update=True)
        click.echo(click.style("Successfully applied ec2nodeclass", bold=True, fg="green"))
    except Exception as e:
        raise e


import click

from tensorkube.constants import get_cluster_name, get_job_sidecar_iam_role_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.services.dynamodb_service import get_dynamodb_access_policy_name
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.iam_service import attach_role_policy, \
    create_or_update_iam_role_with_service_account_cluster_access
from tensorkube.services.job_queue_service import get_job_sidecar_service_account_name, get_job_sidecar_k8s_role_name, \
    get_job_sidecar_rolebinding_name
from tensorkube.services.k8s_service import patch_service_account, create_service_account, create_role_binding
from tensorkube.services.s3_access_service import get_s3_access_policy_name, update_s3_policy
from tensorkube.services.sqs_service import get_sqs_access_policy_name


def apply():
    try:
        s3_access_policy_name = get_s3_access_policy_name()
        update_s3_policy(policy_name=s3_access_policy_name)
        dynamodb_access_policy_name = get_dynamodb_access_policy_name()
        sqs_access_policy_name = get_sqs_access_policy_name()
        role_name = get_job_sidecar_iam_role_name()
        job_queue_sidecar_sa_name = get_job_sidecar_service_account_name()
        k8s_role_name = get_job_sidecar_k8s_role_name()
        role_binding_name = get_job_sidecar_rolebinding_name()

        role = create_or_update_iam_role_with_service_account_cluster_access(get_aws_account_id(),
                                                                             get_cluster_oidc_issuer_url(get_cluster_name()),
                                                                             role_name, job_queue_sidecar_sa_name,
                                                                             'keda')
        role_arn = role['Role']['Arn']
        attach_role_policy(account_no=get_aws_account_id(), policy_name=s3_access_policy_name, role_name=role_name)
        attach_role_policy(account_no=get_aws_account_id(), policy_name=dynamodb_access_policy_name, role_name=role_name)
        attach_role_policy(account_no=get_aws_account_id(), policy_name=sqs_access_policy_name, role_name=role_name)
        create_service_account(job_queue_sidecar_sa_name, "keda")
        patch_service_account(role_arn, job_queue_sidecar_sa_name, "keda")
        create_role_binding(rb_name=role_binding_name, role_name=k8s_role_name, sa_name=job_queue_sidecar_sa_name, rb_namespace="keda",
                            sa_namespace="keda")
        click.echo("Successfully migrated to new SA for fine-tuning ISRA")


    except Exception as e:
        raise e



import subprocess
import tempfile
import time
from importlib import resources
from typing import Optional

from kubernetes import config, client

from tensorkube.constants import get_cluster_name, NAMESPACE, DEFAULT_NAMESPACE, KNATIVE_SERVING_NAMESPACE, \
    CONFIG_FEATURES, ADDON_NAME, get_mount_driver_role_name
from tensorkube.services.aws_service import get_bucket_name, get_aws_account_id
from tensorkube.services.eks_service import install_karpenter, delete_eks_addon, create_eks_addon
from tensorkube.services.eksctl_service import delete_nodegroup
from tensorkube.services.filesystem_service import configure_efs, cleanup_filesystem_resources
from tensorkube.services.k8s_service import check_nodes_ready, evict_pods_from_node, get_nodes_not_using_bottlerocket, \
    drain_and_delete_node, delete_pv_using_name, delete_pvc_using_name_and_namespace, create_build_pv_and_pvc, \
    check_pvc_exists_by_name, get_tensorkube_cluster_context_name
from tensorkube.services.karpenter_service import update_ec2_node_class_ami
from tensorkube.services.logging_service import configure_cloudwatch


def create_and_wait_for_nodegroup():
    cluster_name = get_cluster_name()
    nodegroup_name = f"{cluster_name}-ng-bottlerocket"
    # Access the YAML file using importlib.resources
    package_dir = resources.files('tensorkube')
    yaml_file_path = package_dir / 'configurations' / 'bottlerocket_nodegroup.yaml'

    # Read the YAML file content
    with open(yaml_file_path, 'r') as file:
        filedata = file.read()

    # Replace ${CLUSTER_NAME} in the YAML content
    filedata = filedata.replace('${CLUSTER_NAME}', cluster_name)

    # Step 2: Create a temporary file and write the modified content to it
    with tempfile.NamedTemporaryFile(mode='w+', delete=False) as temp_file:
        temp_file.write(filedata)
        temp_file_path = temp_file.name

    # Step 3: Create the nodegroup using eksctl with the temporary file
    create_command = f"eksctl create nodegroup --config-file={temp_file_path}"
    subprocess.run(create_command, shell=True, check=True)
    print(f"Nodegroup {nodegroup_name} creation initiated.")

    # New Step: Wait for the nodegroup to become ready using Kubernetes client
    label_selector = f"eks.amazonaws.com/nodegroup={nodegroup_name}"
    ready = False
    while not ready:
        ready, ready_nodes = check_nodes_ready(label_selector)
        if ready and len(ready_nodes) > 0:
            print(f"All nodes in nodegroup {nodegroup_name} are ready: {ready_nodes}")
        else:
            print(f"Waiting for nodes in nodegroup {nodegroup_name} to become ready...")
            time.sleep(60)  # Check every 60 seconds

    print(f"Nodegroup {nodegroup_name} is ready.")


def wait_for_deployments_to_be_ready(apps_api, namespaces):
    for ns in namespaces:
        deployments = apps_api.list_namespaced_deployment(ns).items
        for deployment in deployments:
            desired_replicas = deployment.spec.replicas
            while True:
                current_deployment = apps_api.read_namespaced_deployment(deployment.metadata.name, ns)
                ready_replicas = current_deployment.status.ready_replicas if current_deployment.status.ready_replicas else 0
                if ready_replicas >= desired_replicas:
                    print(
                        f"Deployment {deployment.metadata.name} in namespace {ns} is ready with {ready_replicas}/{desired_replicas} replicas.")
                    break
                else:
                    print(
                        f"Waiting for deployment {deployment.metadata.name} in namespace {ns} to be ready. Current status: {ready_replicas}/{desired_replicas} replicas.")
                    time.sleep(10)  # Sleep for 10 seconds before checking again


def enable_knative_pv_pvc_capabilities(namespace=KNATIVE_SERVING_NAMESPACE, context_name: Optional[str] = None):
    if not context_name:
        context_name = get_tensorkube_cluster_context_name()
        if not context_name:
            return None
    k8s_api_client = config.new_client_from_config(context=context_name)
    v1 = client.CoreV1Api(k8s_api_client)

    try:
        # Get the existing config-features ConfigMap
        config_map = v1.read_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace)

        # Update the ConfigMap data
        if config_map.data is None:
            config_map.data = {}
            config_map.data['kubernetes.podspec-nodeselector'] = 'enabled'
            config_map.data['kubernetes.podspec-affinity'] = 'enabled'

        config_map.data["kubernetes.podspec-persistent-volume-claim"] = "enabled"
        config_map.data["kubernetes.podspec-persistent-volume-write"] = "enabled"
        config_map.data["kubernetes.containerspec-addcapabilities"] = "enabled"
        config_map.data["kubernetes.podspec-security-context"] = "enabled"

        # Update the ConfigMap
        v1.patch_namespaced_config_map(name=CONFIG_FEATURES, namespace=namespace, body=config_map)
        print(
            f"Successfully enabled node selector, affinity, pv-claim, pv-write and add-capabilities features in {CONFIG_FEATURES} ConfigMap.")
    except client.exceptions.ApiException as e:
        print(f"Exception when updating ConfigMap: {e}")
        raise


def apply():
    # check if the clusername-ng managed nodegroup exists in the cluster
    # if it does then first cordon all the nodes belonging to the clustername-ng managed nodegroup
    # create the new nodegroup with clustername-ng-bottlerocket and wait for it to get ready
    # scale the deployments from istio-system and knative-serving namespaces to 5 so that they move to new nodes despite the pdbs
    # drain the old nodes by ignoring the daemonsets
    # scale down the deployments back from istio-system and knative-serving namespaces to 1
    # check if all the pods across all the namespaces are ready
    # delete the old nodegroup and wait for it to finish
    # update the cluster cli version to this migration's version
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise Exception("No Kubernetes context found. Please ensure that you have a valid kubeconfig file and try again.")
    k8s_api_client = config.new_client_from_config(context=context_name)  # Load kubeconfig file
    
    core_api = client.CoreV1Api(k8s_api_client)
    apps_api = client.AppsV1Api(k8s_api_client)

    # Step 1 & 2: Check and cordon nodes
    old_nodegroup_name = f'{get_cluster_name()}-ng'
    new_nodegroup_name = f'{get_cluster_name()}-ng-bottlerocket'
    nodegroup_label = 'eks.amazonaws.com/nodegroup'

    # Cordon all nodes in the existing nodegroup
    nodes = core_api.list_node().items

    for node in nodes:
        # Check if the node is part of the target node group by examining its labels
        # if node.metadata.labels.get(nodegroup_label) == old_nodegroup_name:
        #     # Cordon the node by setting it to unschedulable
        core_api.patch_node(node.metadata.name, {"spec": {"unschedulable": True}})
        print(f"Cordoned node {node.metadata.name}")

    # Step 3: Create new nodegroup
    create_and_wait_for_nodegroup()
    print(f"Nodegroup {new_nodegroup_name} created")

    # now for the karpenter configuration
    update_ec2_node_class_ami("default", "Bottlerocket")

    # Step 4: Scale deployments
    namespaces = ["istio-system", "knative-serving"]
    for ns in namespaces:
        deployments = apps_api.list_namespaced_deployment(ns).items
        for deployment in deployments:
            replicas = 5 if ns == "knative-serving" else 2
            apps_api.patch_namespaced_deployment_scale(deployment.metadata.name, ns, {"spec": {"replicas": 5}})
            print(f"Scaled up deployment {deployment.metadata.name} in namespace {ns}")

    wait_for_deployments_to_be_ready(apps_api, namespaces)

    configure_cloudwatch()

    # move karpenter to new nodegroup
    install_karpenter()

    # Step 6: Drain old nodes (use kubectl drain or equivalent in Python)
    for node in nodes:
        # Check if the node is part of the target node group by examining its labels
        if node.metadata.labels.get(nodegroup_label) == old_nodegroup_name:
            # Evict all the pods scheduled there
            evict_pods_from_node(node.metadata.name)
            print(f"Evicted all pods from {node.metadata.name}")

    # Step 7: Scale down deployments
    for ns in namespaces:
        deployments = apps_api.list_namespaced_deployment(ns).items
        for deployment in deployments:
            apps_api.patch_namespaced_deployment_scale(deployment.metadata.name, ns, {"spec": {"replicas": 1}})
            print(f"Scaled down deployment {deployment.metadata.name} in namespace {ns}")

    nodes_without_bottlerocket = get_nodes_not_using_bottlerocket("default")
    for node in nodes_without_bottlerocket:
        drain_and_delete_node(node)

    delete_nodegroup(old_nodegroup_name)

    # Update Knative ConfigMap to enable PV and PVC capabilities, and add capabilities to containers
    enable_knative_pv_pvc_capabilities()

    # delete the s3-pvc in kube-system namespace and then delete s3-pv
    delete_pvc_using_name_and_namespace("s3-claim", NAMESPACE)
    delete_pv_using_name("s3-pv")
    delete_eks_addon(get_cluster_name(), ADDON_NAME)
    # recreate the s3-pv and s3-pvc
    bucket_name = get_bucket_name()
    create_eks_addon(get_cluster_name(), ADDON_NAME, get_aws_account_id(),
                     get_mount_driver_role_name(get_cluster_name()))
    create_build_pv_and_pvc(bucket_name)

    # now check if efs-claim exists
    if check_pvc_exists_by_name(claim_name='efs-pvc', namespace='default'):
        cleanup_filesystem_resources()
    configure_efs()



import subprocess
from pkg_resources import resource_filename
from tensorkube.services.k8s_service import patch_service_account
from tensorkube.constants import get_cluster_name
from tensorkube.services.aws_service import get_aws_account_id
from tensorkube.constants import NODEPOOL, EC2NODECLASS
from tensorkube.services.eks_service import get_cluster_oidc_issuer_url
from tensorkube.services.iam_service import create_iam_role_with_service_account_cluster_access, attach_role_policy, delete_iam_role
from tensorkube.services.karpenter_service import apply_ec2nodeclass, apply_nodepools
from tensorkube.services.nodepool import get_gpu_nodepool, get_gpu_ec2nodeclass, get_config_file, gpu_nodepool_version


def install_nydus():
    apply_ec2nodeclass(get_config_file(EC2NODECLASS, gpu_nodepool_version()), get_gpu_ec2nodeclass())
    apply_nodepools(get_config_file(NODEPOOL, gpu_nodepool_version()), get_gpu_nodepool())  
    install_nydus_snapshotter_helm(get_gpu_nodepool())
    create_nydus_ecr_service_account()
    
    
def delete_nydus():
    uninstall_nydus_snapshotter_helm()
    delete_iam_role(get_nydus_ecr_iam_role_name())


def install_nydus_snapshotter_helm(nodepool_name: str):
    chart_path = resource_filename("tensorkube",'configurations/helm-charts/nydus-snapshotter')
    namespace = 'nydus-snapshotter'
    try:
        subprocess.run(["helm", "upgrade", "--install", "nydus-snapshotter",  chart_path, "--create-namespace", "--namespace", namespace, "--set", f"nodeSelector.nodepool={nodepool_name}"], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Failed to install nydus snapshotter helm chart in namespace '{namespace}' from {chart_path}: {e}")        

def uninstall_nydus_snapshotter_helm():
    try:
        subprocess.run(["helm", "uninstall", "nydus-snapshotter", "--namespace", "nydus-snapshotter"], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Failed to uninstall nydus snapshotter helm chart in namespace 'nydus-snapshotter': {e}")

def get_nydus_runtime_class_name():
    return "runc-nydus"

def get_nydus_ecr_iam_role_name():
    return f"{get_cluster_name()}-nydus-ecr-role"

def get_nydus_snapshoter_name():
    return "nydus-snapshotter"

def get_nydus_snapshoter_namespace():
    return get_nydus_snapshoter_name()

def get_nydus_snapshoter_service_account_name():
    return f"{get_nydus_snapshoter_name()}-sa"

def create_nydus_ecr_service_account():
    namespace = get_nydus_snapshoter_namespace()
    service_account_name = get_nydus_snapshoter_service_account_name()
    role_name = get_nydus_ecr_iam_role_name()
    role = create_iam_role_with_service_account_cluster_access(get_aws_account_id(),get_cluster_oidc_issuer_url(get_cluster_name()), role_name, service_account_name, namespace)
    role_arn = role["Role"]["Arn"]
    attach_role_policy("aws", "AmazonEC2ContainerRegistryReadOnly", role_name)
    patch_service_account(role_arn, service_account_name, namespace)




def get_nydus_image_url(image_url: str):
    return f'{image_url}-nydus'



import click
import importlib.metadata

from tensorkube.constants import get_cluster_name, get_template_bucket_name, DEFAULT_NAMESPACE, \
    get_mount_policy_name, get_mount_driver_role_name, ADDON_NAME, get_cfn_base_stack_name, get_templates_version, \
    AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION, EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION, get_image_registry_id
from tensorkube.helpers import create_mountpoint_driver_role_with_policy
from tensorkube.migration_service.migration_manager.migration_service import set_current_cli_version_to_cluster
from tensorkube.services.aws_service import get_eks_control_plane_available_zones_for_region, \
    get_aws_account_id, get_credentials, get_session_region
from tensorkube.services.build import get_buildkit_service_account_name, configure_buildkit_irsa
from tensorkube.services.cloudformation_service import is_existing_tk_macro, create_cloudformation_stack, \
    stream_stack_events, cloudformation, update_generic_cloudformation_stack
from tensorkube.services.eks_service import install_karpenter, apply_knative_crds, apply_knative_core, \
    apply_nvidia_plugin, create_eks_addon
from tensorkube.services.eksctl_service import create_base_tensorkube_cluster_eksctl
from tensorkube.services.filesystem_service import configure_efs
from tensorkube.services.iam_service import create_mountpoint_iam_policy
from tensorkube.services.istio import check_and_install_istioctl, install_istio_on_cluster, install_net_istio
from tensorkube.services.job_queue_service import create_cloud_resources_for_queued_job_support, \
    create_sa_role_rb_for_job_sidecar
from tensorkube.services.k8s_service import create_aws_secret, create_build_pv_and_pvc
from tensorkube.services.karpenter_service import apply_karpenter_configuration
from tensorkube.services.knative_service import enable_knative_selectors_pv_pvc_capabilities
from tensorkube.services.local_service import check_and_install_cli_tools
from tensorkube.services.logging_service import configure_cloudwatch
from tensorkube.services.nydus import get_nydus_snapshoter_namespace, get_nydus_snapshoter_service_account_name, \
    install_nydus
from tensorkube.services.s3_access_service import create_s3_access_to_pods
from tensorkube.services.s3_service import create_s3_bucket, get_bucket_name


def cfn_configure(test: bool):
    click.echo("Configuring with cloudformation...")
    cluster_name = get_cluster_name()
    stack_name = get_cfn_base_stack_name()
    create_macro = True

    template_bucket_name = get_template_bucket_name(test)

    if test:
        click.echo("Test run. Checking if Macro exists...")
        create_macro = not is_existing_tk_macro()
        click.echo("Check complete")

    zones = get_eks_control_plane_available_zones_for_region()
    parameters = [{"ParameterKey": "ClusterName", 'ParameterValue': cluster_name},
                  {"ParameterKey": "CliVersion", 'ParameterValue': importlib.metadata.version("tensorkube")},
                  {"ParameterKey": "KedaTrainBucketName",
                   'ParameterValue': get_bucket_name(env_name='keda', type="train")},
                  {"ParameterKey": "DefaultEnv", 'ParameterValue': DEFAULT_NAMESPACE},
                  {"ParameterKey": "DefaultEnvBuildBucketName", 'ParameterValue': get_bucket_name()},
                  {"ParameterKey": "DefaultEnvBuildkitISRAServiceAccountName",
                   "ParameterValue": get_buildkit_service_account_name(namespace=DEFAULT_NAMESPACE)},
                  {"ParameterKey": "NydusNamespace", 'ParameterValue': get_nydus_snapshoter_namespace()},
                  {"ParameterKey": "NydusServiceAccountName",
                   'ParameterValue': get_nydus_snapshoter_service_account_name()},
                  {"ParameterKey": "CreateMacro", 'ParameterValue': str(create_macro)},
                  {"ParameterKey": "TemplateBucketName", "ParameterValue": template_bucket_name},
                  {"ParameterKey": "TemplatesVersion", "ParameterValue": get_templates_version()},
                  {"ParameterKey": "AWSAccessLambdaFunctionImageVersion", "ParameterValue": AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION},
                  {"ParameterKey": "EksAccessLambdaFunctionImageVersion", "ParameterValue": EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION},
                  {"ParameterKey": "ImageRegistryId", "ParameterValue": get_image_registry_id(test)},
                  {"ParameterKey": "KedaBuildBucketName", "ParameterValue": get_bucket_name(env_name='keda')},
                  {"ParameterKey": "KedaBuildkitISRAServiceAccountName",
                   "ParameterValue": get_buildkit_service_account_name(namespace='keda')},
                   {"ParameterKey": "Region","ParameterValue": get_session_region()},
                  ]
    for zone in zones:
        parameters.append({"ParameterKey": ("zone"+zone[-1]).upper(), 'ParameterValue': "true"})

    capabilities = ["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"]
    creation_queued, in_created_state = create_cloudformation_stack(
        template_file_path="configurations/cloudformation/configure/tensorkube-base-stack.yaml",
        stack_name=stack_name,
        parameters=parameters,
        capabilities=capabilities)
    if creation_queued:
        stream_stack_events(stack_name)


def legacy_configure(vpc_public_subnets: str, vpc_private_subnets: str):
    if not check_and_install_cli_tools():
        return False
    # TODO!: add helm annotations

    # create cloudformation stack
    cloudformation()

    # create eks cluster
    vpc_public_subnets_list = vpc_public_subnets.split(",") if vpc_public_subnets else []
    vpc_private_subnets_list = vpc_private_subnets.split(",") if vpc_private_subnets else []
    create_base_tensorkube_cluster_eksctl(cluster_name=get_cluster_name(), vpc_public_subnets=vpc_public_subnets_list,
                                          vpc_private_subnets=vpc_private_subnets_list)
    # install karpenter
    install_karpenter()
    # # apply karpenter configuration
    apply_karpenter_configuration()
    configure_cloudwatch()
    #
    # install istio networking plane
    check_and_install_istioctl()
    install_istio_on_cluster()

    # install knative crds
    apply_knative_crds()
    # install knative core
    apply_knative_core()

    # install nvidia plugin
    apply_nvidia_plugin()
    #
    # install net istio
    install_net_istio()

    # create s3 bucket for build
    bucket_name = get_bucket_name()
    create_s3_bucket(bucket_name)

    # create mountpoint policy to mount bucket to eks cluster
    create_mountpoint_iam_policy(get_mount_policy_name(get_cluster_name()), bucket_name)

    # create s3 csi driver role and attach mountpoint policy to it
    create_mountpoint_driver_role_with_policy(cluster_name=get_cluster_name(), account_no=get_aws_account_id(),
                                              role_name=get_mount_driver_role_name(get_cluster_name()),
                                              policy_name=get_mount_policy_name(get_cluster_name()))

    # create eks addon to mount s3 bucket to eks cluster
    create_eks_addon(get_cluster_name(), ADDON_NAME, get_aws_account_id(),
                     get_mount_driver_role_name(get_cluster_name()))

    # create aws credentials cluster secret
    # TODO!: figure out how to update credentials in case of token expiry
    create_aws_secret(get_credentials())

    # create pv and pvc claims for build
    create_build_pv_and_pvc(bucket_name)

    # update knative to use pod labels
    enable_knative_selectors_pv_pvc_capabilities()

    # enable Network files system for the cluster
    click.echo("Configuring EFS for the cluster...")
    configure_efs()

    create_s3_access_to_pods()

    # install keda, create related resources
    create_cloud_resources_for_queued_job_support()
    create_sa_role_rb_for_job_sidecar()

    # configure buildkit irsa, so that pods can access ecr
    configure_buildkit_irsa()

    install_nydus()

    # set current cli version to the cluster
    set_current_cli_version_to_cluster()

    return True


def update_cfn_configure_stack(updated_parameters, test=False):
    click.echo("Updating base cloudformation stack...")
    cluster_name = get_cluster_name()
    stack_name = get_cfn_base_stack_name()
    create_macro = True

    template_bucket_name = get_template_bucket_name(test)

    if test:
        click.echo("Test run. Checking if Macro exists...")
        create_macro = not is_existing_tk_macro()
        click.echo("Check complete")

    zones = get_eks_control_plane_available_zones_for_region()
    parameters = [{"ParameterKey": "ClusterName",
                   'ParameterValue': updated_parameters.get("ClusterName", cluster_name)},
                  {"ParameterKey": "CliVersion",
                   'ParameterValue': updated_parameters.get("CliVersion", importlib.metadata.version("tensorkube"))},
                  {"ParameterKey": "KedaTrainBucketName",
                   'ParameterValue': updated_parameters.get("KedaTrainBucketName", get_bucket_name(env_name='keda', type="train"))},
                  {"ParameterKey": "DefaultEnv",
                   'ParameterValue': updated_parameters.get("DefaultEnv", DEFAULT_NAMESPACE)},
                  {"ParameterKey": "DefaultEnvBuildBucketName",
                   'ParameterValue': updated_parameters.get("DefaultEnvBuildBucketName", get_bucket_name())},
                  {"ParameterKey": "DefaultEnvBuildkitISRAServiceAccountName",
                   "ParameterValue": updated_parameters.get("DefaultEnvBuildkitISRAServiceAccountName",
                                                            get_buildkit_service_account_name(namespace=DEFAULT_NAMESPACE))},
                  {"ParameterKey": "NydusNamespace",
                   'ParameterValue': updated_parameters.get("NydusNamespace", get_nydus_snapshoter_namespace())},
                  {"ParameterKey": "NydusServiceAccountName",
                   'ParameterValue': updated_parameters.get("NydusServiceAccountName",
                                                            get_nydus_snapshoter_service_account_name())},
                  {"ParameterKey": "CreateMacro", 'ParameterValue': str(create_macro)},
                  {"ParameterKey": "TemplateBucketName",
                   "ParameterValue": updated_parameters.get("TemplateBucketName",template_bucket_name)},
                  {"ParameterKey": "TemplatesVersion",
                   "ParameterValue": updated_parameters.get("TemplatesVersion", get_templates_version())},
                  {"ParameterKey": "AWSAccessLambdaFunctionImageVersion",
                   "ParameterValue": updated_parameters.get("AWSAccessLambdaFunctionImageVersion",
                                                            AWS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION)},
                  {"ParameterKey": "EksAccessLambdaFunctionImageVersion",
                   "ParameterValue": updated_parameters.get("EksAccessLambdaFunctionImageVersion",
                                                            EKS_ACCESS_LAMBDA_FUNCTION_IMAGE_VERSION)},
                  {"ParameterKey": "ImageRegistryId",
                   "ParameterValue": updated_parameters.get("ImageRegistryId", get_image_registry_id(test))},
                  {"ParameterKey": "KedaBuildBucketName",
                   "ParameterValue": updated_parameters.get("KedaBuildBucketName",get_bucket_name(env_name='keda'))},
                  {"ParameterKey": "KedaBuildkitISRAServiceAccountName",
                   "ParameterValue": updated_parameters.get("KedaBuildkitISRAServiceAccountName",
                                                            get_buildkit_service_account_name(namespace='keda'))}
                  ]
    for zone in zones:
        parameters.append({"ParameterKey": ("zone" + zone[-1]).upper(), 'ParameterValue': "true"})

    capabilities = ["CAPABILITY_NAMED_IAM", "CAPABILITY_AUTO_EXPAND"]
    update_generic_cloudformation_stack(stack_name=stack_name, parameters=parameters, capabilities=capabilities,
                                        template_file_path='/configurations/cloudformation/configure/tensorkube-base-stack.yaml')



import os
import platform
import shutil
import subprocess
import uuid
from typing import Optional

import boto3
import botocore
import click
import semver

from tensorkube.constants import CliColors
from tensorkube.constants import get_cluster_name, DEFAULT_NAMESPACE, LOCKED_AWS_CLI_VERSION
from tensorkube.services.error import CLIVersionError
import configparser

TENSORFUSE_SESSION = 'tensorfuse'

def get_eks_client():
    session = get_boto3_session()
    return session.client("eks")

def get_session_region():
    session  = get_boto3_session()
    return session.region_name

def get_cloudformation_client():
    session = get_boto3_session()
    return session.client("cloudformation")


def get_ec2_client():
    session = get_boto3_session()
    return session.client("ec2")


def get_iam_client():
    session = get_boto3_session()
    return session.client("iam")

def get_sts_client():
    session = get_boto3_session()
    return session.client("sts")

def get_s3_client():
    session = get_boto3_session()
    return session.client("s3")

def get_s3_resource():
    session = get_boto3_session()
    return session.resource("s3")

def get_ecr_client():
    session = get_boto3_session()
    return session.client("ecr")

def get_efs_client():
    session = get_boto3_session()
    return session.client("efs")


def get_aws_account_id():
    sts = get_sts_client()
    identity = sts.get_caller_identity()
    return identity['Account']

def get_logs_client():
    session = get_boto3_session()
    return session.client("logs")

def get_sqs_client():
    session = get_boto3_session()
    return session.client("sqs")

def get_dynamodb_client():
    session = get_boto3_session()
    return session.client("dynamodb")

def get_dynamodb_resource():
    session = get_boto3_session()
    return session.resource("dynamodb")

def get_aws_user_arn() -> str:
    session = get_boto3_session()
    sts = session.client('sts')
    identity = sts.get_caller_identity()
    return identity['Arn']

def get_boto3_session():
    default_session = boto3.Session()
    config_path = os.path.expanduser(
        os.environ.get("AWS_CONFIG_FILE", "~/.aws/config")
    )
    if TENSORFUSE_SESSION in default_session.available_profiles:
        if not os.path.isfile(config_path):
            text =  f"AWS config file not found at {config_path}; using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        # load and parse
        cfg = configparser.ConfigParser()
        try:
            cfg.read(config_path)
        except Exception as e:
            text = f"Failed to read AWS config file ({e}); using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        profile_section = f"profile {TENSORFUSE_SESSION}"
        try:
            region = cfg.get(profile_section, "region")
        except (configparser.NoSectionError, configparser.NoOptionError) as e:
            text = f"Profile or region not found in config ({e}); using default session"
            click.echo(click.style(text, bold=True, fg=CliColors.WARNING.value))
            return default_session

        # if we got a region, build a session with it
        if region:
            return boto3.Session(profile_name=TENSORFUSE_SESSION, region_name=region)

    return default_session



def get_principal_arn_from_identity_center_arn(arn: str) -> str:
    arn_parts = arn.split(':')
    account_no = arn_parts[4]
    role_name = arn_parts[5].split('/')[1]
    principal_arn = f"arn:aws:iam::{account_no}:role/aws-reserved/sso.amazonaws.com/{role_name}"
    return principal_arn


def get_aws_user_principal_arn() -> str:
    sts = get_sts_client()
    identity = sts.get_caller_identity()
    if 'assumed-role' in identity['Arn']:
        return get_principal_arn_from_identity_center_arn(identity['Arn'])
    return identity['Arn']


def get_karpenter_namespace():
    return "kube-system"


def get_karpenter_version():
    return "0.37.0"


def get_credentials():
    session =  get_boto3_session()
    return session.get_credentials().get_frozen_credentials()


def are_credentials_valid(credentials):
    session = get_boto3_session()
    sts = session.client('sts', aws_access_key_id=credentials.access_key, aws_secret_access_key=credentials.secret_key,
                       aws_session_token=credentials.token)
    try:
        sts.get_caller_identity()
        return True
    except botocore.exceptions.ClientError as e:
        return False



def check_and_install_aws_cli():
    """Check if aws cli is installed and if not install it."""
    try:
        result = subprocess.run(["aws", "--version"], capture_output=True, text=True, check=True)

        # The version is in the format: "aws-cli/2.13.4 Python/3.9.6 Linux/x86_64"
        version_output = result.stdout.strip()
        # Extract the AWS CLI version using string manipulation
        aws_version = semver.VersionInfo.parse(version_output.split(" ")[0].split("/")[1])
        locked_cli_version = semver.VersionInfo.parse(LOCKED_AWS_CLI_VERSION)
        if (aws_version.major, aws_version.minor) < (locked_cli_version.major, locked_cli_version.minor):
            text = f"AWS CLI version is {aws_version}. Please upgrade AWS CLI to version above {LOCKED_AWS_CLI_VERSION}."
            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
            raise CLIVersionError(text)
    except Exception as e:
        if isinstance(e, CLIVersionError):
            raise e
        click.echo(
            click.style("aws-cli not found. Proceeding with installation. Might require sudo password.", bold=True,
                        fg=CliColors.WARNING.value))
        try:
            system = platform.system().lower()
            architecture = platform.machine()
            print(f"Detected operating system: {system.capitalize()}")
            print(f"Detected architecture: {architecture}")
            url = ""
            file_name = ""
            # Define the download URL and file name based on the OS
            if system == "linux":
                url = f"https://awscli.amazonaws.com/awscli-exe-linux-{architecture}-{LOCKED_AWS_CLI_VERSION}.zip"
                file_name = f"awscliv2-{LOCKED_AWS_CLI_VERSION}.zip"
            elif system == "darwin":
                url = f"https://awscli.amazonaws.com/AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
                file_name = f"AWSCLIV2-{LOCKED_AWS_CLI_VERSION}.pkg"
            else:
                print("Unsupported operating system. Please install AWS CLI manually.")
                raise Exception("Unsupported operating system.")
            # Download the installer
            print(f"Downloading AWS CLI installer from {url}...")
            subprocess.run(["curl", "-fsSL", url, "-o", file_name], check=True)
            print("AWS CLI installer downloaded successfully.")
            if system == "linux":
                # Unzip the installer
                print("Unzipping the AWS CLI installer...")
                subprocess.run(["unzip", "-o", file_name], check=True)
                print("AWS CLI installer unzipped successfully.")

                # Install the AWS CLI
                print("Installing AWS CLI...")
                subprocess.run(["sudo", "./aws/install"], check=True)
                print("AWS CLI installed successfully.")

            elif system == "darwin":
                # Use the macOS installer
                print("Installing AWS CLI using macOS package...")
                subprocess.run(["sudo", "installer", "-pkg", file_name, "-target", "/"], check=True)
                print("AWS CLI installed successfully.")
        except subprocess.CalledProcessError as e:
            print(f"Error occurred during installation: {e}")
        finally:
            if os.path.exists(file_name):
                os.remove(file_name)
            if os.path.exists("aws"):
                shutil.rmtree("aws")
            print("Clean-up completed.")

# https://github.com/eksctl-io/eksctl/blob/f4d4076ee22c9026a11a70434b802a7e0f61bd46/pkg/az/az.go#L24
# taken from eksctl source code
zoneIDsToAvoid = {
    "us-east-1": ["use1-az3"],
    "us-west-1": ["usw1-az2"],
    "ca-central-1": ["cac1-az3"],
    "cn-north-1": ["cnn1-az4"],
}


def get_eks_control_plane_available_zones_for_region():
    ec2 = get_ec2_client()
    region = get_session_region()
    response = ec2.describe_availability_zones()
    return [zone['ZoneName'] for zone in response['AvailabilityZones'] if zone['ZoneId'] not in zoneIDsToAvoid.get(region, []) and zone['ZoneType'] == "availability-zone"]



import concurrent.futures
import logging
import mimetypes
import os
import threading
from typing import Optional, NamedTuple, List
import uuid
import boto3
import click
import pathspec
from boto3.s3.transfer import TransferConfig
from tqdm import tqdm

from tensorkube.constants import REGION, get_cluster_name, DEFAULT_NAMESPACE
from tensorkube.services.utils import human_readable_date, human_readable_size
from tensorkube.services.aws_service import get_s3_client, get_s3_resource, get_session_region

class S3ObjectInfo(NamedTuple):
    key: str
    last_modified: str  # Human readable time
    size: str          # Human readable size
    raw_size: int      # Original size in bytes
    is_directory: bool
    content_type: str
    etag: str
    owner: str
    file_extension: str = ''


    @classmethod
    def from_response(cls, obj: dict, head_object: dict = None, is_directory: bool = False):
        """
        Creates an S3ObjectInfo instance from S3 response objects with human-readable formats
        """
        if is_directory:
            return cls(
                key=obj['Prefix'],
                last_modified="Directory",
                size="--",
                raw_size=0,
                is_directory=True,
                content_type="directory",
                etag="--",
                owner="--",
                file_extension=""
            )

        # Get file extension and mime type
        _, file_extension = os.path.splitext(obj['Key'])
        content_type = head_object.get('ContentType', 'application/octet-stream') if head_object else \
                      mimetypes.guess_type(obj['Key'])[0] or 'application/octet-stream'
        return cls(
            key=obj['Key'],
            last_modified=human_readable_date(obj['LastModified']),
            size=human_readable_size(obj['Size']),
            raw_size=obj['Size'],
            is_directory=False,
            content_type=content_type,
            etag=obj.get('ETag', '').strip('"'),
            owner=obj.get('Owner', {}).get('DisplayName', '--'),
            file_extension=file_extension.lower()
        )


# NOTE: bucket name must be universally unique
# NOTE: bucket name should not be guessable
def create_s3_bucket(bucket_name):
    try:
        s3_client = get_s3_client()
        region = get_session_region()
        if region is None or region == REGION:
            s3_client.create_bucket(Bucket=bucket_name)
            click.echo(f"Created bucket {bucket_name} in region {region}")
        else:
            location = {'LocationConstraint': region}
            s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration=location) #TODO: check if this works, can us-west-1 s3 client create a bucket in us-east-1?
    except Exception as e:
        click.echo(f"Error creating bucket {bucket_name} in region {region}: {e}")
        return False

    return True


def list_s3_buckets():
    try:
        s3 = get_s3_client()
        response = s3.list_buckets()
        return response['Buckets']
    except Exception as e:
        click.echo(f"Error listing buckets: {e}")
        return []


# Set up logging for debugging
logging.basicConfig(level=logging.INFO)


def upload_file(file_name, bucket_name, s3_key, s3_client, progress_bar):
    def upload_progress(chunk):
        # logging.debug(f"Uploading chunk of {chunk} bytes for {file_name}")
        progress_bar.update(chunk)

    config = boto3.s3.transfer.TransferConfig(use_threads=False)
    try:
        s3_client.upload_file(file_name, bucket_name, s3_key, Callback=upload_progress,
                              Config=config)  # logging.info(f"Successfully uploaded {file_name} to {bucket_name}/{s3_key}")
    except Exception as e:
        logging.error(f"Error uploading {file_name}: {e}")
    finally:
        progress_bar.close()


def load_dockerignore(file_path):
    with open(file_path, 'r') as f:
        patterns = f.read().splitlines()
    return pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, patterns)


def upload_files_in_parallel(bucket_name, folder_path, s3_path=''):
    s3_client = get_s3_client()
    files_to_upload = []

    dockerignore_filepath = os.path.join(folder_path, '.dockerignore')
    if os.path.exists(dockerignore_filepath):
        spec = load_dockerignore(dockerignore_filepath)
    else:
        spec = pathspec.PathSpec([])

    for root, dirs, files in os.walk(folder_path):
        for file in files:
            local_file = os.path.join(root, file)
            if spec.match_file(local_file):
                continue
            files_to_upload.append(local_file)

    # logging.debug(f"Files to upload: {files_to_upload}")
    progress_bars = {}
    locks = {file: threading.Lock() for file in files_to_upload}

    def upload_with_progress(file_name):
        file_size = os.path.getsize(file_name)
        s3_key = s3_path + os.path.relpath(file_name, folder_path).replace(os.sep, '/')
        # logging.debug(f"Uploading {file_name} to {s3_key} in bucket {bucket_name}")
        with locks[file_name]:
            progress_bars[file_name] = tqdm(total=file_size, unit='B', unit_scale=True, desc=s3_key, leave=True)
        upload_file(file_name, bucket_name, s3_key, s3_client, progress_bars[file_name])

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = {executor.submit(upload_with_progress, file): file for file in files_to_upload}
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logging.error(f"Error in future for {futures[future]}: {e}")

    for progress_bar in progress_bars.values():
        progress_bar.close()


def empty_s3_bucket(bucket_name):
    try:
        s3 = get_s3_resource()
        bucket = s3.Bucket(bucket_name)
        bucket.objects.all().delete()
    except Exception as e:
        click.echo(f"Error emptying bucket {bucket_name}: {e}")
        return False

    return True


def empty_s3_folder(bucket_name, folder_name):
    try:
        s3 = get_s3_resource()
        bucket = s3.Bucket(bucket_name)
        bucket.objects.filter(Prefix=folder_name).delete()
    except Exception as e:
        click.echo(f"Error emptying folder {folder_name} in bucket {bucket_name}: {e}")
        return False

    return True


def delete_s3_bucket(bucket_name):
    try:
        click.echo(f"Emptying bucket {bucket_name}...")
        empty_s3_bucket(bucket_name)
        click.echo(f"Deleting bucket {bucket_name}...")
        s3 = get_s3_client()
        s3.delete_bucket(Bucket=bucket_name)
        click.echo(f"Deleted bucket {bucket_name}")
    except Exception as e:
        click.echo(f"Error deleting bucket {bucket_name}: {e}")
        return False

    return True

def check_if_a_bucket_type_exists(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return True
    return False


def list_bucket_contents(bucket_name: str, prefix: str = '') -> List[S3ObjectInfo]:
    """
    Lists all objects in an S3 bucket with detailed, human-readable information.

    Args:
        bucket_name: Name of the S3 bucket
        prefix: Optional prefix to filter objects (like a directory path)

    Returns:
        List of S3ObjectInfo containing human-readable file/directory information
    """
    try:
        s3_client = get_s3_client()
        paginator = s3_client.get_paginator('list_objects_v2')
        objects = []

        for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix, Delimiter='/'):
            # Handle directories (common prefixes)
            if 'CommonPrefixes' in page:
                for prefix_obj in page['CommonPrefixes']:
                    objects.append(S3ObjectInfo.from_response(prefix_obj, is_directory=True))

            # Handle files
            if 'Contents' in page:
                for obj in page['Contents']:
                    if obj['Key'] == prefix:
                        continue

                    try:
                        head_object = s3_client.head_object(Bucket=bucket_name, Key=obj['Key'])
                    except Exception as _:
                        head_object = None

                    objects.append(S3ObjectInfo.from_response(obj, head_object))

        return sorted(objects, key=lambda x: (x.is_directory, x.key.lower()))

    except Exception as e:
        logging.error(f"Error listing contents of bucket {bucket_name}: {e}")
        return []


def delete_file_from_s3(bucket_name: str, object_name: str) -> bool:
    """
    Deletes a single file from an S3 bucket

    Args:
        bucket_name: Name of the S3 bucket
        object_name: Name of the file to delete

    Returns:
        bool: True if deletion was successful, False otherwise
    """
    try:
        s3_client = get_s3_client()
        s3_client.delete_object(Bucket=bucket_name, Key=object_name)
        return True
    except Exception as e:
        logging.error(f"Error deleting file {object_name} from bucket {bucket_name}: {e}")
        return False


# TODO!: make function generic to get any config value
def get_bucket_name(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return bucket['Name']
    else:
        bucket_name = f"{prefix}{str(uuid.uuid4())[:18]}"
        if len(bucket_name) > 63:
            raise ValueError("Environment name is too long. Please use a shorter name.")
        return bucket_name

def get_existing_bucket_name(env_name: Optional[str] = None, type: str = "build"):
    buckets = list_s3_buckets()
    if env_name and env_name != DEFAULT_NAMESPACE:
        prefix = f'{get_cluster_name()}-{env_name}-{type}-bucket-'
    else:
        prefix = f'{get_cluster_name()}-{type}-bucket-'
    # checks if such a bucket already exists
    for bucket in buckets:
        if bucket['Name'].startswith(prefix):
            return bucket['Name']

    return None


import click
from pkg_resources import resource_filename

from tensorkube.services.k8s_service import get_tensorkube_cluster_context_name, list_all_namespaces
from tensorkube.services.knative_service import list_ksvc_in_namespace, apply_virtual_service_for_routing


def apply():
    # list all ksvcs across all namespaces
    # apply a virtualservice for all those ksvcs
    context_name = get_tensorkube_cluster_context_name()
    if not context_name:
        raise Exception(
            "No Kubernetes context found. Please ensure that you have a valid kubeconfig file and try again.")

    virtual_service_yaml_file_path = resource_filename('tensorkube',
                                                       'configurations/build_configs/virtual_service.yaml')

    all_namespaces = list_all_namespaces()
    for ns in all_namespaces:
        click.echo(f'Applying migration to all services in the {ns} environment')
        ksvcs = list_ksvc_in_namespace(ns)
        for ksvc in ksvcs['items']:
            service_name = ksvc['metadata']['name']
            apply_virtual_service_for_routing(service_name=service_name, yaml_file_path=virtual_service_yaml_file_path,
                sanitised_project_name=service_name.split('-')[0], env=ns, context_name=context_name)
            click.echo(
                f'Migrated routing for service {service_name} in namespace {ns}')
